﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;
using System.ServiceModel;
using System.Collections.ObjectModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;
using System.Diagnostics;
using System.Runtime.CompilerServices;

namespace MVeldhuizen.DevTools.ServiceBehaviors
{
    /// <summary>
    /// Service and/or Endpoint behavior that causes client access policy files to be hosted on the
    /// appropriate address, so that the service can be accessed from Silverlight and Adobe Flash.
    /// 
    /// This behavior extension is intended for use together with WCF Service Host. See 
    /// http://thorarin.net/blog/post.aspx?id=70618cd9-eb8c-4d9f-8e49-9bdde3b2eba6 for details.
    /// </summary>
    public sealed class CrossDomainServiceBehavior : BehaviorExtensionElement, IServiceBehavior, IEndpointBehavior
    {
        private static readonly string[] _uriSchemes = new string[] { "http", "https" };
        private static readonly object _syncRoot = new object();
        private static List<Uri> _listenAddresses = new List<Uri>();
        private static ServiceHost _serviceHost;
        
        #region BehaviorExtensionElement Implementation

        public override Type BehaviorType
        {
            get
            {
                return typeof(CrossDomainServiceBehavior);
            }
        }

        protected override object CreateBehavior()
        {
            return new CrossDomainServiceBehavior();
        }

        #endregion

        #region IServiceBehavior Members

        void IServiceBehavior.AddBindingParameters(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase, Collection<ServiceEndpoint> endpoints, BindingParameterCollection bindingParameters)
        {
            // Ignore IMetadataExchange endpoints
            if (endpoints[0].Contract.ContractType == typeof(IMetadataExchange))
                return;

            HostCrossDomainPolicy(endpoints);
        }

        void IServiceBehavior.ApplyDispatchBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
        }

        void IServiceBehavior.Validate(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase)
        {
        }

        #endregion

        #region IEndpointBehavior Members

        void IEndpointBehavior.AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
            HostCrossDomainPolicy(endpoint);
        }

        void IEndpointBehavior.ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
        }

        void IEndpointBehavior.ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
        }

        void IEndpointBehavior.Validate(ServiceEndpoint endpoint)
        {
        }

        #endregion

        private void HostCrossDomainPolicy(ServiceEndpoint endpoint)
        {
            HostCrossDomainPolicy(new ServiceEndpoint[] { endpoint });
        }

        private static void HostCrossDomainPolicy(IEnumerable<ServiceEndpoint> endpoints)
        {
            lock (_syncRoot)
            {
                var addresses =
                    GetListenAddressesFromEndpoints(endpoints).
                    Where(uri => !_listenAddresses.Contains(uri)).
                    ToArray();

                if (addresses.Length == 0)
                    return;

                _listenAddresses.AddRange(addresses);

                if (_serviceHost != null)
                    _serviceHost.Close();

                _serviceHost = CreateServiceHost(_listenAddresses);
                _serviceHost.Open();
            }
        }

        /// <summary>
        /// Creates a ServiceHost instance to the CrossDomainPolicyService on the specified addresses.
        /// </summary>
        /// <param name="addresses">The addresses to host the policy service on.</param>
        /// <returns>A ServiceHost instance, in unopened state.</returns>
        private static ServiceHost CreateServiceHost(IList<Uri> addresses)
        {
            ServiceHost serviceHost = new ServiceHost(typeof(CrossDomainPolicyService));
            foreach (Uri address in addresses)
            {
                ServiceEndpoint endpoint = serviceHost.AddServiceEndpoint(typeof(ICrossDomainPolicyService), new WebHttpBinding(), address.ToString());
                endpoint.Behaviors.Add(new WebHttpBehavior());
            }

            return serviceHost;
        }

        /// <summary>
        /// Get root addresses for any HTTP and HTTPS endpoints.
        /// </summary>
        private static IEnumerable<Uri> GetListenAddressesFromEndpoints(IEnumerable<ServiceEndpoint> endpoints)
        {
            return endpoints.
                Select(ept => ept.Address.Uri).
                Where(uri => _uriSchemes.Contains(uri.Scheme)).
                Select(uri => new Uri(uri, "/"));
        }
    }
}
