﻿using System;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Description;
using System.ServiceModel.Configuration;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;
using System.Collections.ObjectModel;

namespace MVeldhuizen.DevTools.ServiceBehaviors
{
    /// <summary>
    /// This behavior will make WCF services send SOAP faults with a HTTP 200 status code,
    /// instead of the default HTTP 500 status code. This makes the proper error handing
    /// possible in Silverlight.
    /// See http://thorarin.net/blog/post.aspx?id=52e96e3d-e0c2-466a-887c-84c384648c19 for
    /// a more detailed description.
    /// </summary>
    /// <remarks>
    /// By default, WCF services return soap faults with a HTTP 500 status code. 
    /// Due to limitations in the browser networking API, the bodies of these messages are 
    /// inaccessible within Silverlight, and consequently the fault messages cannot be used.
    /// 
    /// To be able to use FaultExceptions in Silverlight, a WCF service must send the fault
    /// messages with a HTTP 200 status code instead of the HTTP 500 status code. This
    /// change will not impact the way error handling works for regular .NET clients.
    /// </remarks>
    public sealed class SilverlightFaultBehavior : BehaviorExtensionElement, IServiceBehavior, IEndpointBehavior
    {
        public override Type BehaviorType
        {
            get { return typeof(SilverlightFaultBehavior); }
        }

        protected override object CreateBehavior()
        {
            return new SilverlightFaultBehavior();
        }

        #region IServiceBehavior Members

        public void AddBindingParameters(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase, Collection<ServiceEndpoint> endpoints, BindingParameterCollection bindingParameters)
        {
            foreach (ServiceEndpoint endpoint in endpoints)
            {
                if (!endpoint.Behaviors.OfType<SilverlightFaultBehavior>().Any())
                    endpoint.Behaviors.Add(this);
            }
        }

        public void ApplyDispatchBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase) {}
        public void Validate(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase) {}

        #endregion

        #region IEndpointBehavior Members

        void IEndpointBehavior.ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
            endpointDispatcher.DispatchRuntime.MessageInspectors.Add(new ServiceFaultMessageInspector());
        }

        // The following methods are stubs and not relevant. 
        void IEndpointBehavior.AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters) { }
        void IEndpointBehavior.ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime) { }
        void IEndpointBehavior.Validate(ServiceEndpoint endpoint) { }

        #endregion

        /// <summary>
        /// Change status code from HTTP 500 to HTTP 200.
        /// </summary>
        private sealed class ServiceFaultMessageInspector : IDispatchMessageInspector
        {
            void IDispatchMessageInspector.BeforeSendReply(ref Message reply, object correlationState)
            {
                if (reply.IsFault)
                {
                    reply.Properties[HttpResponseMessageProperty.Name] = new HttpResponseMessageProperty()
                                                                         {
                                                                             StatusCode = System.Net.HttpStatusCode.OK
                                                                         };
                }
            }

            object IDispatchMessageInspector.AfterReceiveRequest(ref Message request, IClientChannel channel, InstanceContext instanceContext)
            {
                return null;
            }
        }
    }

}
